/* NSUndoManager.h created by veilljf on Mon 02-Nov-1998 */

#import "NSStackInvocationManager.h"
#import <Foundation/NSArray.h>

@implementation NSStackInvocationManager


/*" Instances Method "*/

- init
{
    id sign;
    undoEventList = [[NSMutableArray alloc] init];
    undoInvocation = 0;
    undoing = NO;

    sign = [self methodSignatureForSelector:@selector(beginInvo)];
    BEGIN_INVOCATION = [[NSInvocation invocationWithMethodSignature:sign] retain];
    END_INVOCATION = [[NSInvocation invocationWithMethodSignature:sign] retain];
    [BEGIN_INVOCATION setTarget:self];
    [END_INVOCATION setTarget:self];
    [BEGIN_INVOCATION setSelector:@selector(beginInvo)];
    [END_INVOCATION setSelector:@selector(endInvo)];


    return self;
}

- (void)dealloc
{
    [undoEventList release];
    [BEGIN_INVOCATION release];
    [END_INVOCATION release];
    [super dealloc];
}

- (int)levelsOfUndo
{
    return levelsOfUndo;
}

- (void)setLevelsOfUndo:(int)l
{
    levelsOfUndo = l;
}

- (void)runInLoop:(id)arg
/*" This is the method that is getting called "after delay" so that it is at the begining of an event loop.  This is to create a 'fake' hook into the event loop so that we enclose all undo events into the same group of events that are whithin one seconds.  Keep getting called from the event loop.  This method close the previous group and open a new one.

    Than call itself after a delay of one seconds. "*/
{
    [NSObject cancelPreviousPerformRequestsWithTarget:self
                                         selector:@selector(runInLoop:)
                                           object:nil];

    // indicate that we are running in event loop mode
    runInLoop = YES;

    // recall myself after one seconds.
    [self performSelector:@selector(runInLoop:)
               withObject:nil
               afterDelay:1.0];

    // close the previous grouping
    [self endUndoGrouping];
    // then start a new group for the new loop.
    [self beginUndoGrouping];
    
}

- (void)beginInvo
    /*" Begin  event Group"*/
{
    undoInvocation--;
}

- (void)endInvo
    /*" Ending event Group"*/
{
    undoInvocation++;
}

- (void)registerUndoWithTarget:tar
                     selector:(SEL)sel
                       object:ob
    /*" Register an undo invocation."*/
{
    if(undoing) {
        // actualy when register method while undoing we should register those as redo.
    } else {
        NSMethodSignature *msign = [tar methodSignatureForSelector:sel];
        NSInvocation* invo = [NSInvocation invocationWithMethodSignature:msign];
        [invo setSelector:sel];
        [invo setTarget:tar];
        [invo setArgument:&ob atIndex:2];
        if([invo target] != self) {
            [invo retainArguments];
        }
        [self addEvent:invo];
    }
}

- (void)addEvent:invo
    /*"Add the next NSInvocation to the event array."*/
{
    [undoEventList addObject:invo];
}

- (BOOL)isUndoing
    /*"Am I currently in an undo operation"*/
{
    return (undoInvocation > 0);
}

- (void)removeAllActions
/*"clear the undo stack"*/
{
    [undoEventList removeAllObjects];
}

- (BOOL)canUndo
{
    return ([undoEventList count] != 0);
}

- (void)undo
    /*" Execute last elements of the event stack until you reach the end of the current group."*/
{
    if([self canUndo]) {
        id lastEvent;
        undoing = YES;
        lastEvent = [undoEventList lastObject];
        do {
            [lastEvent invoke];
            [undoEventList removeLastObject];
            lastEvent = [undoEventList lastObject];
        } while(!(([lastEvent isEqual:BEGIN_INVOCATION]) && ( undoInvocation == 1)));
        [lastEvent invoke];
        [undoEventList removeLastObject];

        undoing = NO;
    }
}

- (void)beginUndoGrouping
    /*"register the begining of an undo group"*/
{
    [self addEvent:BEGIN_INVOCATION];
}

- (void)endUndoGrouping
    /*"register the ending of an undo group"*/
{
    if([[undoEventList lastObject] isEqual:BEGIN_INVOCATION]) {
        [undoEventList removeLastObject];
    } else {
        [self addEvent:END_INVOCATION];
    }
}

- (void)setActionName:(NSString*)name
    /*"Register the name of the current group"*/
{
    [self registerUndoWithTarget:self
                        selector:@selector(doNothingButHoldValue:)
                            object:name];
}

- (void)doNothingButHoldValue:(NSString*)value
{
    // do nothing
}

- (id) initWithCoder:(NSCoder *)aDecoder
    /*"
    Unarchives a stack instance.
     "*/
{
    undoEventList = [[aDecoder decodeObject] retain];
    [aDecoder decodeValueOfObjCType:@encode(BOOL) at:&runInLoop];
    BEGIN_INVOCATION = [[aDecoder decodeObject] retain];
    END_INVOCATION = [[aDecoder decodeObject] retain];

    undoInvocation = 0;
    undoing = NO;

    return self;
}

- (void) encodeWithCoder:(NSCoder *)aCoder
    /*"
    Archives a stack instance.
     "*/
{
    [aCoder encodeObject:undoEventList];
    [aCoder encodeValueOfObjCType:@encode(BOOL) at:&runInLoop];
    [aCoder encodeObject:BEGIN_INVOCATION];
    [aCoder encodeObject:END_INVOCATION];
}

@end

@implementation NSStackInvocationManager (myStepAdditions)
/*" This category add AppKit context where we have  menu cells interactions. "*/

- (NSString*)lastGroupName
/*""*/
{
    id result = nil;
    BOOL done = NO;
    if([self canUndo]) {
        id lastEvent;
        int idx = [undoEventList count];
        int level = 0;
        while( (idx > 0) && (done == NO) ) {
            lastEvent = [undoEventList objectAtIndex:idx - 1];
            if(!strcmp((char*)[lastEvent selector], "doNothingButHoldValue:" )) {
                [lastEvent getArgument:&result atIndex:2];
                done = YES;
            }
            if (!strcmp((char*)[lastEvent selector],"beginInvo")) {
                level --;
                if (level == 0)  {
                    // do not go over previous undo group
                    done = YES;
                }
            }
            if(!strcmp((char*)[lastEvent selector], "endInvo")) {
                level ++;
            }
            idx --;
        }
    }
    return result;
}

@end